#pragma once

#include <mfapi.h>
#include <mfidl.h>
#include <mfreadwrite.h>

#include <list>



#define MAXVideoURLLenght                   256

#define VFL_OK                              0
#define VFL_ERROR                           -1
#define VFL_NO_FILE_URL_SET                 1
#define VFL_URL_TO_LONG                     2
#define VFL_PATH_TO_LONG                    3
#define VFL_NO_VIDEO_SOURCE                 4
#define VFL_VIDEO_ATTRIBUTES_NOT_SET        5
#define VFL_VIDEO_BUFFER_IS_NULL            6
#define VFL_COULD_NOT_CONVERT_PATH_TO_URL   25
#define VFL_COULD_NOT_READ_SOURCE_FILE      26
#define VFL_COULD_NOT_DECODE_FILE           27
#define VFL_COULD_NOT_GET_MEDIA_TYPE        28
#define VFL_COULD_NOT_GET_MEDIA_ATTRIBUTES  29
#define VFL_COULD_NOT_READ_ANY_FRAMES       30
#define VFL_NO_MORE_FRAMES                  31

typedef int VFLResult;

using namespace std;

//! This class loads video files
/*!
    This class can load video files into a BYTE-Array. It uses the windows media foundation framework
    and should work under Windows 7, 8 and later. It can load all frames of a video file, all frames
    after an offset or single frames with a given distance to the last loaded frame.
 */

class CVideoFrameLoader
{
public:
    //! This is the default constructor.
    /*!
       It simple initialises the object. To open a video file you can use the other constructor, the
       setVideoPath(PWSTR filePath) method or the setVideoURL(PWSTR fileURL) method.
       \sa CVideoFrameLoader(PWSTR filePath), setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL)
     */
	CVideoFrameLoader(void);

    //! Constructor to load a file.
    /*!
      Initialises the object and loads the specified file.
      \param filePath a pointer to a WSTR which contains the path to the video file in windows syntax
      \sa CVideoFrameLoader(), setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL)
     */
	CVideoFrameLoader(PWSTR filePath);

    //! Destructor
    /*!
      Cleans up all reserved memory.
     */
	~CVideoFrameLoader(void);

    //! Opens a new videofile.
    /*!
      If there is another video file loaded, the file will be closed and all
      references will be deleted.
      \param filePath a pointer to a WSTR which contains the path to the video file in windows syntax
      \return returns VFL_OK if the operation succeeded.
      \sa CVideoFrameLoader(PWSTR filePath), setVideoURL(PWSTR fileURL)
     */
    VFLResult setVideoPath(PWSTR filePath);

    //! Opens a new videofile.
    /*!
      If there is another video file loaded, the file will be closed and all
      references will be deleted.
      \param fileURL a pointer to a WSTR which contains the URL to the video file
      \return returns VFL_OK if the operation succeeded
      \sa CVideoFrameLoader(PWSTR filePath), setVideoPath(PWSTR filePath)
     */
    VFLResult setVideoURL(PWSTR fileURL);

    //! Loads all video frames of the opend video file.
    /*!
      The frames will be saved into a BYTE buffer. The buffer can then be retrieved
      by calling getVideoBuffer(PBYTE *videoArray).
      \return returns VFL_NO_MORE_FRAMES if the operation succeded and the video has ended.
      \sa loadAllFramesIntoBuffer(int offset, int distanceBetween), loadAllFramesIntoBuffer(int offset, int distanceBetween, int lastFrame)
      \sa addNextFrameToBuffer(), addNextFrameToBuffer(int distanceFromLast)
      \sa setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL), CVideoFrameLoader(PWSTR filePath)
     */
    VFLResult loadAllFramesIntoBuffer();

    //! Loads all video frames with offset and distance between.
    /*!
      The frames will be saved into a BYTE buffer. The buffer can then be retrieved
      by calling getVideoBuffer(PBYTE *videoArray).
      \param offset an integer value starting by 0 that specifies the offset of frames from the start of the video
      \param distanceBetween an integer value the specifies the number of frames between to frames, that are left out
      \return returns VFL_NO_MORE_FRAMES if the operation succeded and the video has ended.
      \sa loadAllFramesIntoBuffer(), loadAllFramesIntoBuffer(int offset, int distanceBetween, int lastFrame)
      \sa addNextFrameToBuffer(), addNextFrameToBuffer(int distanceFromLast)
      \sa setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL), CVideoFrameLoader(PWSTR filePath)
     */
    VFLResult loadAllFramesIntoBuffer(int offset, int distanceBetween);

    //! Loads all video frames with offset, distance between and sets a last frame
    /*!
      The frames will be saved into a BYTE buffer. The buffer can then be retrieved
      by calling getVideoBuffer(PBYTE *videoArray).
      \param offset an integer value starting by 0 that specifies the offset of frames from the start of the video
      \param distanceBetween an integer value the specifies the number of frames between to frames, that are left out
      \param lastFrame an inteber value specifying the last frame that should be read; if 0 the method reads till the last frame of the video file
      \return returns VFL_OK if the operation succeded or VFL_NO_MORE_FRAMES if the operation succeded and the video has ended
      \sa loadAllFramesIntoBuffer(), loadAllFramesIntoBuffer(int offset, int distanceBetween)
      \sa addNextFrameToBuffer(), addNextFrameToBuffer(int distanceFromLast)
      \sa setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL), CVideoFrameLoader(PWSTR filePath)
     */
    VFLResult loadAllFramesIntoBuffer(int offset, int distanceBetween, int lastFrame);

    //! Adds the next frame at the current video file position
    /*!
      After that, the "play head" moves to the next frame.
      \return returns VFL_OK if the operation succeded or VFL_NO_MORE_FRAMES if the "play head" is at the end of the file
      \sa addNextFrameToBuffer(int distanceFromLast)
      \sa loadAllFramesIntoBuffer(), loadAllFramesIntoBuffer(int offset, int distanceBetween), loadAllFramesIntoBuffer(int offset, int distanceBetween, int lastFrame)
      \sa setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL), CVideoFrameLoader(PWSTR filePath)
     */
    VFLResult addNextFrameToBuffer();

    //! Adds the next frame at an offset from the current video file position
    /*!
      The function skipps the specified number of frames in the video file and reads the frame at that position
      \param distanceFromLast integer value that specifies the distance from the last read frame; if 0 the next consecutive frame is read
      \return returns VFL_OK if the operation succeded or VFL_NO_MORE_FRAMES if the "play head" comes to the end of the file without reading a frame
      \sa addNextFrameToBuffer()
      \sa loadAllFramesIntoBuffer(), loadAllFramesIntoBuffer(int offset, int distanceBetween), loadAllFramesIntoBuffer(int offset, int distanceBetween, int lastFrame)
      \sa setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL), CVideoFrameLoader(PWSTR filePath)
     */
    VFLResult addNextFrameToBuffer(int distanceFromLast);

    //! Returns the resolution of the current loaded video file
    /*!
      This method returns good data, only after the video file has been succesfully loaded.
      \param width pointer to an integer, where the video width will be saved
      \param height pointer to an integer, where the video height will be saved
      \return return VFL_OK if the video resulution could be obtained
      \sa setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL), CVideoFrameLoader(PWSTR filePath)
     */
    VFLResult getVideoResolution(int *width, int *height);

    //! Returns the BYTE-Array in which the video data is stored.
    /*!
      The caller receives a pointer to the byte array in which the video data resides in BGRA-format.
      The size of the array is the number of frames times the width and the height of the video times 4.
      These values can be obtained by calling the apropriate methods. After this method has been called,
      the array will never be delete. It falls to the caller to delete the array after usage.
      \param videoArray a pointer to a pointer to a BYTE
      \return VFL_OK if the call was successful
      \sa getVideoResolution(int *width, int *height), getBufferFrameCount(int *frameCount)
      \sa addNextFrameToBuffer(), addNextFrameToBuffer(int distanceFromLast)
      \sa loadAllFramesIntoBuffer(), loadAllFramesIntoBuffer(int offset, int distanceBetween), loadAllFramesIntoBuffer(int offset, int distanceBetween, int lastFrame)
      \sa setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL), CVideoFrameLoader(PWSTR filePath)
     */
    VFLResult getVideoBuffer(PBYTE *videoArray);

    //! Returns the number of fames that have been saved to the byte array.
    /*!
      \param frameCount a pointer to an integer where the number of captured frames will be saved
      \return returns VFL_OK if the operation was successful
      \sa getVideoBuffer(PBYTE *videoArray), getVideoResolution(int *width, int *height)
      \sa addNextFrameToBuffer(), addNextFrameToBuffer(int distanceFromLast)
      \sa loadAllFramesIntoBuffer(), loadAllFramesIntoBuffer(int offset, int distanceBetween), loadAllFramesIntoBuffer(int offset, int distanceBetween, int lastFrame)
     */
    VFLResult getBufferFrameCount(int *frameCount);

    //! Resets the "play head" to the starting position.
    /*!
      The byte array will remain filled. To completly reset the object you have to reopen the video file.
      \sa setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL), CVideoFrameLoader(PWSTR filePath)
     */
    void reset();

    //! Sets the position of the "play head" to a new time code.
    /*!
      The byte array will remain filled. To completly reset the object you have to reopen the video file.
      \param hnsPosition a LONGLONG value of the new "play head" position in 100 ms increments.
      \return returns VFL_OK if the operation was successful
      \sa addNextFrameToBuffer(), addNextFrameToBuffer(int distanceFromLast)
      \sa loadAllFramesIntoBuffer(), loadAllFramesIntoBuffer(int offset, int distanceBetween), loadAllFramesIntoBuffer(int offset, int distanceBetween, int lastFrame)
     */
    HRESULT setPosition(const LONGLONG& hnsPosition);
protected:
    PBYTE p_videoArray; /*!< Pointer to the video array. It's only filled when the getVideoBuffer(PBYTE *videoArray) or the getVideoBuffer(PBYTE *videoArray) method is called. */

    //! Completly resets the object.
    /*!
      This virtual method is called to reset the reset all neccessary member properties bevor a new video file is loaded.
      Overwrite this method to reset your own properties before a new video file is loaded, if you derive from this class.
      \return returns VFL_OK;
      \sa reset()
      \sa setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL), CVideoFrameLoader(PWSTR filePath)
     */
    virtual VFLResult resetProperties();

    //! Copies the video data from the internal storage into the video array.
    /*!
      Call this method in your derived class if you want to access the data in a consecutive memory block.
      After the call the p_videoArray pointer points to a byte array with the video data. If only this method is
      invoked, the byte array will be destroyed bevor a new video is loaded or the object is deleted.
      \sa getVideoBuffer(PBYTE *videoArray)
      \sa addNextFrameToBuffer(), addNextFrameToBuffer(int distanceFromLast)
      \sa loadAllFramesIntoBuffer(), loadAllFramesIntoBuffer(int offset, int distanceBetween), loadAllFramesIntoBuffer(int offset, int distanceBetween, int lastFrame)
     */
    void copyVideoListToVideoArray();
private:
    PWSTR p_videoURL;
    IMFSourceReader *p_videoReader;
    int m_frameCount;
    int m_totalFrameCount;
    UINT32 m_videoWidth, m_videoHeight;
    bool m_videoBufferTransfered;
    bool m_comInitialisationSuccessful;
    bool m_mediaFoundationInitalisationSuccessful;

    list<PBYTE, allocator<PBYTE>> *p_videoList;

    void initializeFrameLoader();
    VFLResult setVideoPath(PWSTR filePath, bool reset);
    VFLResult setVideoURL(PWSTR fileURL, bool reset);
    VFLResult openFileInSourceReader();
    HRESULT configureDecoder(IMFSourceReader *pReader);
    VFLResult copyFrameToBuffer(IMFSample *pSample);
    void SafeRelease(IUnknown **ppT);
};

