#pragma once

#include "VideoFrameLoader.h"

#include <QGLWidget>

//! This class loads video files and save them into a 3D texture.
/*!
    This class can load video files into a BYTE-Array and transfer this array into a 3D texture.
    It is derived from the CVideoFrameLoader class.
    It uses the windows media foundation framework and should work under Windows 7, 8 and later.
    It can load all frames of a video file, all frames after an offset or single frames with a given distance
    to the last loaded frame.
 */

class CVideoFrameToTextureLoader : public CVideoFrameLoader
{
public:
    //! This is the default constructor.
    /*!
       It simple initialises the object. To open a video file you can use the other constructor, the
       setVideoPath(PWSTR filePath) method or the setVideoURL(PWSTR fileURL) method.
       \sa CVideoFrameToTextureLoader(PWSTR filePath), setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL)
    */
    CVideoFrameToTextureLoader();

    //! Constructor to load a file.
    /*!
      Initialises the object and loads the specified file.
      \param filePath a pointer to a WSTR which contains the path to the video file in windows syntax
      \sa CVideoFrameToTextureLoader(), setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL)
     */
    CVideoFrameToTextureLoader(PWSTR filePath);

    //! Destructor
    /*!
      Cleans up all reserved memory.
     */
    ~CVideoFrameToTextureLoader(void);

    //! Loads all video frames of the opend video file and transfers it into the bound texture.
    /*!
      The frames will be saved into a BYTE buffer. The buffer is then transfered into a 3D texture.
      You have to bind an apropriate texture before calling this method.
      \return returns VFL_NO_MORE_FRAMES if the operation succeded and the video has ended.
      \sa loadAllFramesIntoTexture(int offset, int distanceBetween), loadAllFramesIntoTexture(int offset, int distanceBetween, int lastFrame)
      \sa setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL), CVideoFrameToTextureLoader(PWSTR filePath)
      \sa transferVideoBufferToTexture()
     */
    VFLResult loadAllFramesIntoTexture();

    //! Loads all video frames with offset and distance between and transfers it into the bound texture.
    /*!
      The frames will be saved into a BYTE buffer. The buffer is then transfered into a 3D texture.
      You have to bind an apropriate texture before calling this method.
      \param offset an integer value starting by 0 that specifies the offset of frames from the start of the video
      \param distanceBetween an integer value the specifies the number of frames between to frames, that are left out
      \return returns VFL_NO_MORE_FRAMES if the operation succeded and the video has ended.
      \sa loadAllFramesIntoTexture(), loadAllFramesIntoTexture(int offset, int distanceBetween, int lastFrame)
      \sa setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL), CVideoFrameToTextureLoader(PWSTR filePath)
      \sa transferVideoBufferToTexture()
     */
    VFLResult loadAllFramesIntoTexture(int offset, int distanceBetween);

    //! Loads all video frames with offset, distance between and sets a last frame and transfers it into the bound texture
    /*!
      The frames will be saved into a BYTE buffer. The buffer is then transfered into a 3D texture.
      You have to bind an apropriate texture before calling this method.
      \param offset an integer value starting by 0 that specifies the offset of frames from the start of the video
      \param distanceBetween an integer value the specifies the number of frames between to frames, that are left out
      \param lastFrame an inteber value specifying the last frame that should be read; if 0 the method reads till the last frame of the video file
      \return returns VFL_OK if the operation succeded or VFL_NO_MORE_FRAMES if the operation succeded and the video has ended
      \sa loadAllFramesIntoTexture(), loadAllFramesIntoTexture(int offset, int distanceBetween)
      \sa setVideoPath(PWSTR filePath), setVideoURL(PWSTR fileURL), CVideoFrameToTextureLoader(PWSTR filePath)
      \sa transferVideoBufferToTexture()
     */
    VFLResult loadAllFramesIntoTexture(int offset, int distanceBetween, int lastFrame);

    //VFLResult addNextFrameToTexture();
    //VFLResult addNextFrameToTexture(int distanceFromLast);

    //! Transfers the video buffer byte array into a previously bound texture.
    /*!
      The video array is transfered into a 3D texture.
      You have to bind an apropriate texture before calling this method.
      \return returns VFL_OK if the operation succeeded
      \sa loadAllFramesIntoBuffer(), loadAllFramesIntoBuffer(int offset, int distanceBetween), loadAllFramesIntoBuffer(int offset, int distanceBetween, int lastFrame)
      \sa addNextFrameToBuffer(), addNextFrameToBuffer(int distanceFromLast)
     */
    VFLResult transferVideoBufferToTexture();

protected:
    virtual VFLResult resetProperties();
    PFNGLTEXIMAGE3DPROC glTexImage3D;
};

