#ifndef GLWIDGET_H
#define GLWIDGET_H




#include <QOpenGLBuffer>
#include <QOpenGLShader>
#include <QOpenGLFunctions>
#include <QOpenGLFramebufferObject>

#include <QGLWidget>
#include <QWidget>

#include "VideoFrameToTextureLoader.h"

#define VV_DRAWMODE_VOLUME      0
#define VV_DRAWMODE_VIDEODIFF   1

#define VV_METRIC_NONE          0
#define VV_METRIC_IQDIF         1
#define VV_METRIC_YDIF          2
#define VV_METRIC_YNDIF         3
#define VV_METRIC_YMSE          4
#define VV_METRIC_YNMSE         5

typedef uint VVMetric;

//! This class loads video files and renders a volume visualisation using OpenGL 3.3.
/*!
    The volume visualisation is constructed using a ray casting algorithm written in GLSL.
    The result can be modiefied using the following methods. It is derived from QGLWidget and QOpenGLFunctions
    and should be inserted as a widget into a QT program.
 */

class GLWidget : public QGLWidget, protected QOpenGLFunctions//, protected QGLFunctions
{
    Q_OBJECT

public:
    //! This is the default constructor
    /*!
      QT uses this constructor of you put the widget into a Qt designer form
      \param parent a pointer to the parent widget
     */
    GLWidget(QWidget *parent = 0);

    //! This is the default destructor
    ~GLWidget();

    QSize minimumSizeHint() const;
    QSize sizeHint() const;

    //! Opens a video file
    /*!
      The video file should be in a format, that is readable by a codec installed on the target machine.
      The widget is using the custom CVideoFrameToTextureLoader class to open the video file.
      It can also be called to change to video file.
      \param videoFilePath a QString containing the path to the video file in windows format
      \param hRes a pointer to an integer receiving the horizontal resolution of the loaded video
      \param vRes a pointer to an integer receiving the vertical resolution of the loaded video
      \return returns VFL_OK if the operation completed successfully
      \sa loadVideoFile(int startFrame, int stepSize, int endFrame)
     */
    VFLResult openVideoFile(const QString& videoFilePath, int *hRes, int *vRes);

    //! Loads the currently open video file and displays it.
    /*!
      The video file is loaded into a 3D Texture. This method can also be called to reload the file.
      When this method is called the old video data is deleted.
      \param startFrame integer that specifies the first frame to load (starting with 1)
      \param stepSize integer that specifies the steps between frames (starting with 1)
      \param endFrame integer that specifies the last frame, if 0 the method tries to load all frames in the file
      \return returns VFL_OK if the operation completed successfully
      \sa openVideoFile(const QString& videoFilePath, int *hRes, int *vRes)
     */
    VFLResult loadVideoFile(int startFrame, int stepSize, int endFrame);

    //! Returns the number of frames that have been loaded
    /*!
      \return an integer with the number of frames that have been loaded
      \sa openVideoFile(const QString& videoFilePath, int *hRes, int *vRes), loadVideoFile(int startFrame, int stepSize, int endFrame)
     */
    int getFrameCount();

    //! Changes the offset of the displayed video
    /*!
      The frame offset changes the first frame that is visible on the bounding value.
      All other frames are rotated accordingly. When the frame offset is set, a redraw is initiated.
      \param frameOffset integer that specifies the start frame (starts at 0)
     */
    void setFrameOffset(int frameOffset);

    //! Sets the style of the displayed video volume
    /*!
      There are only two options. Horse show view or block view. When this method is called, a redraw is initiated.
      \param horseShoeView boolean that specifies the style; if true the horse shoe view is selected, block view is otherwise chosen
      \sa setBoxSize(int boxSize)
     */
    void setStyle(bool horseShoeView);

    //! Sets the scale factor of the volume box in z-Direction
    /*!
      This value is saved regardless of the set style. When the box style is chosen, the change becomes visible.
      Nonetheless, the method initiates a redraw
      \param boxSize an integer specifying the scale factor
      \sa setStyle(bool horseShoeView)
     */
    void setBoxSize(int boxSize);

    //! Sets the stepsize used by the ray casting algorithm
    /*!
      Use this method with caution. Very small values can overtax the graphics card easyly.
      This method initiates a redraw.
      \param stepSize float value specifying the step size (0 < stepSize < 1)
     */
    void setStepSize(float stepSize);

    //! Sets the used metric
    /*!
      The avaliable metrices are VV_METRIC_NONE, VV_METRIC_YDIF, VV_METRIC_YMSE, VV_METRIC_YNDIF, VV_METRIC_YNMSE or VV_METRIC_IQDIF.
      Depending on the metric and if it's shown first the method can take a long time to calculate the neccessary prerequisits for
      the volume visualisation. This method initiates a redraw.
      \param metric VVMetric value that specifies the desired metric
     */
    void setMetric(VVMetric metric);

    //! Sets the used transfer function
    /*!
      This parameter has to be set to get any good result. The transfer function must be an evenly spaced aray of
      4 component float vectors. The components are RGBA. The values are distributed over the luminace axis and specify,
      the opacity (A) over the complete range of the luminace. For all the metrics accept VV_METRIC_NONE the color values are
      also used to define the color of the voxels with this luminance. The vectors in between are interpolated.
      \param tfArray pointer to a float array
      \param arraySize number of 4-component float vectors
      \sa setMetric(VVMetric metric)
     */
    void setTransferFunction(GLfloat *tfArray, int arraySize);

protected:
    void initializeGL();
    void paintGL();
    void resizeGL(int width, int height);
    void mouseMoveEvent(QMouseEvent *event);
    void mousePressEvent(QMouseEvent *event);
    void mouseReleaseEvent(QMouseEvent *event);
    void wheelEvent(QWheelEvent *event);

    PFNGLGENVERTEXARRAYSPROC glGenVertexArrays;
    PFNGLBINDVERTEXARRAYPROC glBindVertexArray;
    PFNGLTEXIMAGE3DPROC glTexImage3D;
private:
    bool loadShaders();
    bool prepareShaderProgram( const QString& vertexShaderPath,const QString& fragmentShaderPath, QOpenGLShaderProgram **shader);
    void createBufferObjects();

    void setupCamera();
    void setupModelMatrix();

    void calculateVideoDiffBuffer(bool normalized);
    void calculateMeanAndStdDev();

    QOpenGLShaderProgram *m_volumeShader;
    QOpenGLShaderProgram *m_videoDiffShader;
    QOpenGLBuffer *m_boxPosVBO;
    QOpenGLBuffer *m_texCoordVBO;
    QOpenGLBuffer *m_boxIBO;
    QOpenGLBuffer *m_squarePosVBO;
    QOpenGLBuffer *m_squareTexCoordVBO;
    QOpenGLBuffer *m_squareIBO;

    GLuint m_videoTexture;
    GLuint m_transferTexture;
    GLuint m_meanTexture;
    GLuint m_varTexture;
    GLuint m_boxVAO;
    GLuint m_squareVAO;

    QMatrix4x4 m_viewMatrix;
    QMatrix4x4 m_projectionMatrix;
    QMatrix4x4 m_modelMatrix;

    CVideoFrameToTextureLoader *p_videoFrameToTextureLoader;
    PBYTE p_videoBuffer;
    PBYTE p_videoDiffBuffer;
    PBYTE p_videoNDiffBuffer;
   // GLfloat *pVideoDiffMeanBuffer;
   // GLfloat *p_videoDiffDeviationBuffer;

    int m_videoWidth;
    int m_videoHeight;
    int m_frameCount;

    uint m_drawMode;

    float m_frameOffset;
    float m_cameraAngleUp;
    float m_cameraAngleSide;
    float m_cameraDistance;

    QVector3D m_cameraCenter;
    QVector3D m_backgroundColor;
    QVector3D m_cameraPosition;

    GLfloat *p_transferFunction;

    bool m_horseShowView;
    float m_stepSize;
    int m_boxSize;

    VVMetric m_metric;

    QPoint m_oldMousePos;
};

#endif // GLWIDGET_H
